
/* ====================================================================
 * Copyright (c) 1995 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission.
 *
 * 5. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */


/*
 * http_trans.c: Stuff for dealing with directory transes
 * 
 * Original by Rob McCool, rewritten in succession by David Robinson
 * and rst.
 * 
 */

#include "httpd.h"
#include "http_config.h"

#include "libtrans.h"

#include <stdarg.h>

typedef struct {
    array_header *tabs;
    array_header *langs;
    table *charsets;
    table *agents;
} trans_dir_conf;

module trans_module;

/**************** Configuration handling *********************/

static void *create_trans_config (pool *p, char *dummy)
{
    trans_dir_conf *a =
      (trans_dir_conf *)pcalloc (p, sizeof(trans_dir_conf));

    a->tabs = make_array (p, 4, sizeof(char*));
    a->langs = make_array (p, 4, sizeof(char*));
    a->charsets = make_table (p, 4);
    a->agents = make_table (p, 4);
    return a;
}

static void *merge_trans_config (pool *p, void *basev, void *addv)
{
    trans_dir_conf *a =
	(trans_dir_conf *)pcalloc (p, sizeof(trans_dir_conf));
    trans_dir_conf *base = (trans_dir_conf *)basev,
	*add = (trans_dir_conf *)addv;

    a->tabs = append_arrays (p, add->tabs, base->tabs);
    a->langs = append_arrays (p, add->langs, base->langs);
    a->charsets = overlay_tables (p, add->charsets, base->charsets);
    a->agents = overlay_tables (p, add->agents, base->agents);
    return a;
}

static char *add_trantab(cmd_parms *cmd, trans_dir_conf *s, char *l)
{
    char **a=(char**)push_array(s->tabs);
    *a=pstrdup(cmd->pool,l);
    return NULL;
}

static char *add_tranlang(cmd_parms *cmd, trans_dir_conf *s, char *l)
{
    char **a=(char**)push_array(s->langs);
    *a=pstrdup(cmd->pool,l);
    return NULL;
}

static char *add_trancs(cmd_parms *cmd, trans_dir_conf *s, char *f, char *r)
{
    table_set (s->charsets,r,f);
    return NULL;
}

static char *add_tranua(cmd_parms *cmd, trans_dir_conf *s, char *f, char *r)
{
    table_set (s->agents,r,f);
    return NULL;
}

static command_rec trans_cmds[] = {
{ "TranslateTable", add_trantab, NULL, OR_FILEINFO, ITERATE, 
    "valid translation table names"},
{ "TranslateLanguage", add_tranlang, NULL, OR_FILEINFO, ITERATE, 
    "languages that should be translated"},
{ "TranslateCharset", add_trancs, NULL, OR_FILEINFO, TAKE2, 
    "translation name and charset name"},
{ "TranslateUserAgent", add_tranua, NULL, OR_FILEINFO, ITERATE2, 
    "translation name and user agent substrings"},
{ NULL }
};

/***************** Translation functions ********************/

static long trans_send_fd(FILE *f, request_rec *r)
{
    char buf[IOBUFSIZE];
    long total_bytes_sent;
    register int n,o,w;
    conn_rec *c = r->connection;
    int i;

    total_bytes_sent = 0;
    while (!r->connection->aborted) {
        while ((n= fread(buf, sizeof(char), IOBUFSIZE, f)) < 1
	       && ferror(f) && errno == EINTR)
	    continue;
	
	if (n < 1) {
            break;
        }
        o=0;

	for (i=0;i<n;i++) buf[i]=TR_OUT(buf[i]);

	if (strncmp(buf,"#!/usr/local/bin/htmltrans",26) == 0) {
	    o+=26;
	    n-=26;
	}
        if (r->bytes_sent != -1) r->bytes_sent += n;
	total_bytes_sent += n;

        while(n && !r->connection->aborted) {
#ifdef APACHE_SSL
            w=os_conn_write(&buf[o],sizeof(char),n,c);
#else
            w=fwrite(&buf[o],sizeof(char),n,c->client);
#endif
	    if (w)
	        reset_timeout(r); /* reset timeout after successfule write */
            n-=w;
            o+=w;
        }
    }
#ifdef APACHE_SSL
    os_conn_flush(c);
#else
    fflush(c->client);
#endif
    
    return total_bytes_sent;
}

static int trans_rputc (int c, request_rec *r)
{
    if (r->connection->aborted) return EOF;
#ifdef APACHE_SSL
    os_conn_putc (c, r->connection);
#else
    putc (TR_OUT(c), r->connection->client);
#endif
    ++r->bytes_sent;
    return c;
}

static long trans_rprintf (request_rec *r, char *fmt, ...)
{
    va_list args;
    int retval,i;
    static char tmp[MAX_STRING_LEN];
    
    if (r->connection->aborted) return EOF;
    
    va_start (args, fmt);
    retval = vsprintf (tmp, fmt, args);
    va_end (args);

    for (i=0;i<retval;i++) tmp[i]=TR_OUT(tmp[i]);
#ifdef APACHE_SSL
    retval = os_conn_write(tmp,sizeof(char),retval,r->connection);
#else
    retval = fputs (tmp, r->connection->client);
#endif
    r->bytes_sent += retval;
    return retval;
}

extern long (*fn_send_fd)(FILE *f, request_rec *r);
extern int (*fn_rputc)(int c, request_rec *r);
extern long (*fn_rprintf)(request_rec *r, char *fmt, ...);

static void initialize_trans (server_rec *s, pool *p) {

    fn_send_fd = trans_send_fd;
    fn_rputc = trans_rputc;
    fn_rprintf = trans_rprintf;
    return;
}

/****************** Translation setup *********************/

int setup_translation (request_rec *r)
{
    char trname[4],csparm[32],*p;
    int i,do_translate=0;
    table *e = r->subprocess_env;
    trans_dir_conf *conf =
	(trans_dir_conf *)get_module_config(r->per_dir_config, &trans_module);
    array_header *tabs=conf->tabs;
    array_header *langs=conf->langs;
    table *charsets=conf->charsets;
    table *agents=conf->agents;
    char *acs=table_get(r->headers_in,"Accept-charset");
#ifdef TRANS_DEBUG
    printf ("check: %s \"%s\",\"%s\" and reset (%s)\n",
		r->server->server_hostname,
		r->content_language?r->content_language:"(NULL)",
		r->content_type?r->content_type:"(NULL)",
		r->filename?r->filename:"(NULL)");
#endif
    /* reset translation */
    settrtab_byname("nul");
    trname[0]='\0';

    if (r->content_language) {
	char **ll=(char**)langs->elts;
	for (i=0;i<langs->nelts;i++) {
	    if (strcasecmp(ll[i],r->content_language) == 0) {
#ifdef TRANS_DEBUG
		printf("allowed language: %s found\n",r->content_language);
#endif
		do_translate=1;
		break;
	    }
	}
    } else {
	/* kludge - by local conventions rhtmls and rcgis are Russian */
	if ((strcmp(r->filename+strlen(r->filename)-4,"rcgi") == 0) ||
	    (strcmp(r->filename+strlen(r->filename)-5,"rhtml") == 0) ||
	    (strcmp(r->filename+strlen(r->filename)-5,"xhtml") == 0)) {
	    /* r->content_language=pstrdup(r->pool,"ru"); */
	    do_translate=1;
	}
    }

    if (!do_translate) {
#ifdef TRANS_DEBUG
	printf("No translation for \"%s\",\"%s\"\n",
		r->content_language?r->content_language:"(null)",
		r->content_type?r->content_type:"(null)");
#endif
	return DECLINED;
    }

    do_translate=0;

    if (acs) {
#ifdef TRANS_DEBUG
	printf("accept-charset: \"%s\"\n",acs);
#endif
	if (p=table_get(conf->charsets,acs)) {
	    strncpy(trname,p,3);
	    trname[3]='\0';
	    do_translate=1;
	}
    }

    if (!do_translate) {
	char **ll=(char**)tabs->elts;
	strncpy(trname,r->server->server_hostname,3);
	trname[3]='\0';
	for (i=0;i<tabs->nelts;i++) {
#ifdef TRANS_DEBUG
	    printf("check allowed name \"%s\"\n",ll[i]);
#endif
	    if (strcasecmp(ll[i],trname) == 0) {
#ifdef TRANS_DEBUG
		printf("server name matches allowed table \"%s\"\n",ll[i]);
#endif
		do_translate=1;
		break;
	    }
	}
    }

    if (!do_translate) {
	char *ua=table_get(r->headers_in,"User-Agent");
	if (ua) {
	    table_entry *ll=(table_entry*)agents->elts;
#ifdef TRANS_DEBUG
	    printf("check user-agent \"%s\"\n",ua);
#endif
	    for (i=0;i<agents->nelts;i++) {
		if (strstr(ua,ll[i].key)) {
#ifdef TRANS_DEBUG
		    printf("user agent matches allowed table \"%s\"\n",
				ll[i].key);
#endif
		    do_translate=1;
		    strncpy(trname,ll[i].val,3);
		    trname[3]='\0';
		    break;
		}
	    }
	}
    }

    if (do_translate) {
	table_entry *ll=(table_entry*)charsets->elts;
	for (i=0;i<charsets->nelts;i++) {
	    if (strcasecmp(trname,ll[i].val) == 0) {
#ifdef TRANS_DEBUG
		printf("adding charset \"%s\"\n",ll[i].key);
#endif
		r->content_charset=pstrdup(r->pool,ll[i].key);
		break;
	    }
	}
	table_set (e, "LANG",trname);
	table_set (e, "CHARSET_TRANSLATION",trname);
#ifdef TRANS_DEBUG
	printf("use \"%s\"\n",trname);
#endif
	i=settrtab_byname(trname);
#ifdef TRANS_DEBUG
	printf("settrtab(%s) rc=%d\n",trname,i);
#endif
    }

    return DECLINED;
}

module trans_module = {
   STANDARD_MODULE_STUFF,
   initialize_trans,		/* initializer */
   create_trans_config,		/* dir config creater */
   merge_trans_config,		/* dir merger --- default is to override */
   NULL,			/* server config */
   NULL,			/* merge server configs */
   trans_cmds,			/* command table */
   NULL,			/* handlers */
   NULL,			/* filename translation */
   NULL,			/* check_user_id */
   NULL,			/* check auth */
   NULL,			/* check access */
   NULL,			/* type_checker */
   setup_translation,		/* fixups */
   NULL				/* logger */
};
